@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -22,6 +22,8 @@
 # This file contains classes that handle layout of displayables on
 # the screen.
 
+from __future__ import print_function
+
 from renpy.display.render import render, Render
 import renpy.display
 
@@ -118,6 +120,9 @@ class Container(renpy.display.core.Displayable):
 
     def _duplicate(self, args):
 
+        if args and args.args:
+            args.extraneous()
+
         if not self._duplicatable:
             return self
 
@@ -259,8 +264,9 @@ class Container(renpy.display.core.Displayable):
         return False
 
 
-def LiveComposite(size, *args, **properties):
+def Composite(size, *args, **properties):
     """
+    :name: Composite
     :doc: disp_imagelike
 
     This creates a new displayable of `size`, by compositing other
@@ -276,7 +282,7 @@ def LiveComposite(size, *args, **properties):
 
     ::
 
-       image eileen composite = LiveComposite(
+       image eileen composite = Composite(
            (300, 600),
            (0, 0), "body.png",
            (0, 0), "clothes.png",
@@ -299,8 +305,13 @@ def LiveComposite(size, *args, **properties):
     return rv
 
 
+LiveComposite = Composite
+
+
 class Position(Container):
     """
+    :undocumented:
+
     Controls the placement of a displayable on the screen, using
     supplied position properties. This is the non-curried form of
     Position, which should be used when the user has directly created
@@ -773,6 +784,7 @@ class MultiBox(Container):
         spacings = [ first_spacing ] + [ spacing ] * (len(self.children) - 1)
 
         box_wrap = self.style.box_wrap
+        box_wrap_spacing = self.style.box_wrap_spacing
         xfill = self.style.xfill
         yfill = self.style.yfill
         xminimum = self.style.xminimum
@@ -874,7 +886,7 @@ class MultiBox(Container):
                 if box_wrap and remwidth - sw - padding < 0 and line:
                     maxx, maxy = layout_line(line, target_width - x, 0)
 
-                    y += line_height
+                    y += line_height + box_wrap_spacing
                     x = 0
                     line_height = 0
                     remwidth = width
@@ -916,7 +928,7 @@ class MultiBox(Container):
                 if box_wrap and remheight - sh - padding < 0:
                     maxx, maxy = layout_line(line, 0, target_height - y)
 
-                    x += line_width
+                    x += line_width + box_wrap_spacing
                     y = 0
                     line_width = 0
                     remheight = height
@@ -989,6 +1001,10 @@ class MultiBox(Container):
 
         except IgnoreLayers:
             if self.layers:
+
+                if ev.type != renpy.display.core.TIMEEVENT:
+                    renpy.display.interface.post_time_event()
+
                 return None
             else:
                 raise
@@ -1022,7 +1038,7 @@ class SizeGroup(renpy.object.Object):
         maxwidth = 0
 
         for i in self.members:
-            rend = i.render(width, height, st, at)
+            rend = renpy.display.render.render_for_size(i, width, height, st, at)
             maxwidth = max(rend.width, maxwidth)
 
         self._width = maxwidth
@@ -1048,7 +1064,9 @@ class Window(Container):
             self.add(child)
 
     def visit(self):
-        return [ self.style.background ] + self.children
+        rv = [ ]
+        self.style._visit_window(rv.append)
+        return rv + self.children
 
     def get_child(self):
         return self.style.child or self.child
@@ -1063,10 +1081,6 @@ class Window(Container):
 
             group.members.append(self)
 
-    def predict_one(self):
-        pd = renpy.display.predict.displayable
-        self.style._predict_window(pd)
-
     def render(self, width, height, st, at):
 
         # save some typing.
@@ -1082,6 +1096,9 @@ class Window(Container):
         if size_group and size_group in size_groups:
             xminimum = max(xminimum, size_groups[size_group].width(width, height, st, at))
 
+        width = max(xminimum, width)
+        height = max(yminimum, height)
+
         left_margin = scale(style.left_margin, width)
         left_padding = scale(style.left_padding, width)
 
@@ -1263,8 +1280,8 @@ class DynamicDisplayable(renpy.display.core.Displayable):
 
                 for i in child:
                     renpy.display.predict.displayable(i)
-                else:
-                    renpy.display.predict.displayable(child)
+            else:
+                renpy.display.predict.displayable(child)
 
         except:
             pass
@@ -1322,6 +1339,7 @@ def condition_switch_predict(switch, predict_all=None):
 
 def ConditionSwitch(*args, **kwargs):
     """
+    :name: ConditionSwitch
     :doc: disp_dynamic
     :args: (*args, predict_all=None, **properties)
 
@@ -1450,19 +1468,23 @@ class IgnoresEvents(Container):
         return None
 
 
-def LiveCrop(rect, child, **properties):
+def Crop(rect, child, **properties):
     """
     :doc: disp_imagelike
+    :name: Crop
 
-    This created a displayable by cropping `child` to `rect`, where
+    This creates a displayable by cropping `child` to `rect`, where
     `rect` is an (x, y, width, height) tuple. ::
 
-        image eileen cropped = LiveCrop((0, 0, 300, 300), "eileen happy")
+        image eileen cropped = Crop((0, 0, 300, 300), "eileen happy")
     """
 
     return renpy.display.motion.Transform(child, crop=rect, **properties)
 
 
+LiveCrop = Crop
+
+
 class Side(Container):
 
     possible_positions = set([ 'tl', 't', 'tr', 'r', 'br', 'b', 'bl', 'l', 'c'])
@@ -1501,8 +1523,14 @@ class Side(Container):
         super(Side, self)._clear()
         self.sized = False
 
+    def per_interact(self):
+        self.sized = False
+
     def render(self, width, height, st, at):
 
+        if renpy.config.developer and len(self.positions) != len(self.children):
+            raise Exception("A side has the wrong number of children.")
+
         pos_d = { }
         pos_i = { }
 
@@ -1545,10 +1573,8 @@ class Side(Container):
                 if pos not in pos_d:
                     return owidth, oheight
 
-                rend = render(pos_d[pos], width, height, st, at)
-                rv = max(owidth, rend.width), max(oheight, rend.height)
-                rend.kill()
-                return rv
+                rend = renpy.display.render.render_for_size(pos_d[pos], width, height, st, at)
+                return max(owidth, rend.width), max(oheight, rend.height)
 
             cwidth, cheight = sizeit('c', width, height, 0, 0)
             cwidth, top = sizeit('t', cwidth, height, cwidth, top)
@@ -1619,17 +1645,35 @@ class Side(Container):
         row2 = top + tops
         row3 = top + tops + cheight + bottoms
 
-        place('c', col2, row2, cwidth, cheight)
+        place_order = [
+            ('c', col2, row2, cwidth, cheight),
 
-        place('t', col2, row1, cwidth, top)
-        place('r', col3, row2, right, cheight)
-        place('b', col2, row3, cwidth, bottom)
-        place('l', col1, row2, left, cheight)
+            ('t', col2, row1, cwidth, top),
+            ('r', col3, row2, right, cheight),
+            ('b', col2, row3, cwidth, bottom),
+            ('l', col1, row2, left, cheight),
 
-        place('tl', col1, row1, left, top)
-        place('tr', col3, row1, right, top)
-        place('br', col3, row3, right, bottom)
-        place('bl', col1, row3, left, bottom)
+            ('tl', col1, row1, left, top),
+            ('tr', col3, row1, right, top),
+            ('br', col3, row3, right, bottom),
+            ('bl', col1, row3, left, bottom),
+        ]
+
+        # This sorts the children for placement according to
+        # their order in positions.
+        if renpy.config.keep_side_render_order:
+            def sort(elem):
+                pos, x, y, w, h = elem
+
+                if pos not in pos_d:
+                    return
+
+                return self.positions.index(pos)
+
+            place_order.sort(key=sort)
+
+        for pos, x, y, w, h in place_order:
+            place(pos, x, y, w, h)
 
         return rv
 
@@ -1739,15 +1783,16 @@ class AdjustTimes(Container):
         return self.child.get_placement()
 
 
-class LiveTile(Container):
+class Tile(Container):
     """
     :doc: disp_imagelike
+    :name: Tile
 
     Tiles `child` until it fills the area allocated to this displayable.
 
     ::
 
-        image bg tile = LiveTile("bg.png")
+        image bg tile = Tile("bg.png")
 
     """
 
@@ -1783,6 +1828,9 @@ class LiveTile(Container):
         return rv
 
 
+LiveTile = Tile
+
+
 class Flatten(Container):
     """
     :doc: disp_imagelike
@@ -1821,6 +1869,8 @@ class Flatten(Container):
 
         return rv
 
+    def get_placement(self):
+        return self.child.get_placement()
 
 class AlphaMask(Container):
     """
@@ -1833,11 +1883,7 @@ class AlphaMask(Container):
     opaque where `child` and `mask` are both opaque.
 
     The `child` and `mask` parameters may be arbitrary displayables. The
-    size of the AlphaMask is the size of the overlap between `child` and
-    `mask`.
-
-    Note that this takes different arguments from :func:`im.AlphaMask`,
-    which uses the mask's color channel.
+    size of the AlphaMask is the size of `child`.
     """
 
     def __init__(self, child, mask, **properties):
@@ -1846,24 +1892,19 @@ class AlphaMask(Container):
         self.add(child)
         self.mask = renpy.easy.displayable(mask)
         self.null = None
-        self.size = None
 
     def render(self, width, height, st, at):
 
         cr = renpy.display.render.render(self.child, width, height, st, at)
-        mr = renpy.display.render.render(self.mask, width, height, st, at)
+        w, h = cr.get_size()
 
-        cw, ch = cr.get_size()
-        mw, mh = mr.get_size()
+        mr = renpy.display.render.Render(w, h)
+        mr.place(self.mask, main=False)
 
-        w = min(cw, mw)
-        h = min(ch, mh)
-        size = (w, h)
+        if self.null is None:
+            self.null = Fixed()
 
-        if self.size != size:
-            self.null = Null(w, h)
-
-        nr = renpy.display.render.render(self.null, width, height, st, at)
+        nr = renpy.display.render.render(self.null, w, h, st, at)
 
         rv = renpy.display.render.Render(w, h, opaque=False)
 
