@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2019 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -181,6 +181,11 @@ def walkdir(dir):  # @ReservedAssignment
         if i[0] == ".":
             continue
 
+        try:
+            i = renpy.exports.fsdecode(i)
+        except:
+            continue
+
         if os.path.isdir(dir + "/" + i):
             for fn in walkdir(dir + "/" + i):
                 rv.append(i + "/" + fn)
@@ -460,6 +465,14 @@ def load_core(name):
         if rv is not None:
             return rv
 
+    # Look for the file directly.
+    if not renpy.config.force_archives:
+        try:
+            fn = transfn(name)
+            return open_file(fn, "rb")
+        except:
+            pass
+
     # Look for the file in the apk.
     for apk in apks:
         prefixed_name = "/".join("x-" + i for i in name.split("/"))
@@ -469,14 +482,6 @@ def load_core(name):
         except IOError:
             pass
 
-    # Look for the file directly.
-    if not renpy.config.force_archives:
-        try:
-            fn = transfn(name)
-            return open_file(fn, "rb")
-        except:
-            pass
-
     # Look for it in archive files.
     for prefix, index in archives:
         if not name in index:
@@ -514,6 +519,22 @@ def load_core(name):
     return None
 
 
+def check_name(name):
+    """
+    Checks the name to see if it violates any of Ren'Py's rules.
+    """
+
+    if renpy.config.reject_backslash and "\\" in name:
+        raise Exception("Backslash in filename, use '/' instead: %r" % name)
+
+    if renpy.config.reject_relative:
+
+        split = name.split("/")
+
+        if ("." in split) or (".." in split):
+            raise Exception("Filenames may not contain relative directories like '.' and '..': %r" % name)
+
+
 def get_prefixes(tl=True):
     """
     Returns a list of prefixes to search for files.
@@ -565,11 +586,6 @@ def loadable_core(name):
     if name in loadable_cache:
         return loadable_cache[name]
 
-    for apk in apks:
-        prefixed_name = "/".join("x-" + i for i in name.split("/"))
-        if prefixed_name in apk.info:
-            return True
-
     try:
         transfn(name)
         loadable_cache[name] = True
@@ -577,6 +593,12 @@ def loadable_core(name):
     except:
         pass
 
+    for apk in apks:
+        prefixed_name = "/".join("x-" + i for i in name.split("/"))
+        if prefixed_name in apk.info:
+            loadable_cache[name] = True
+            return True
+
     for _prefix, index in archives:
         if name in index:
             loadable_cache[name] = True
@@ -722,7 +744,7 @@ class RenpyImporter(object):
         code = compile(source, filename, 'exec', renpy.python.old_compile_flags, 1)
         exec code in mod.__dict__
 
-        return mod
+        return sys.modules[fullname]
 
     def get_data(self, filename):
         return load(filename).read()
